﻿// Photoshop FITS Liberator - Concatenator
// Copyright 2006 ESA/ESO/NASA
// All Rights Reserved

// -----------------------------------------------------------------------------
// Behavior - Defines how the Concatenator tool merges metadata. Users are free 
// to edit the properties below to tweak Concatenator's behavior.
// -----------------------------------------------------------------------------

#script "The ESA/ESO/NASA FITS Concatenator"
#target photoshop
#strict on

var Behavior = {
    /** Should the processed documents be closed? */
    CloseDocuments      : true,
    /** Should the processed documents be saved, and if so, how? 
        See SaveOptions in the documentation for more information. */
    SaveOption          : SaveOptions.DONOTSAVECHANGES,
    /** Default bitdepth to use if the proper bitdepth cannot be determined. */
    DefaultBitDepth     : BitsPerChannelType.EIGHT,
    /** Composite (merged) document name. */
    CompositeName       : "Composite",
    /** Composite (merged document) resolution. */
    DefaultResolution   : 72,
    /** Should documents be flattened before merging? */
    MergeFlattens       : true,
    /** Sentinel used for missing exposure values. */
    Sentinel            : "-",
    /** Is debugging mode on? */
    Debug               : false
};

// -----------------------------------------------------------------------------
// Includes 
// -----------------------------------------------------------------------------

#include "xmlsax.inc"
#include "xmlw3cdom.inc"

// -----------------------------------------------------------------------------
// Internal constants
// -----------------------------------------------------------------------------

/** Keyword types. */
var Keyword = {
    Simple  : 0,
    Complex : 1,
    Seq     : 2,
    Alt     : 3,
    Bag     : 4
};

/** Which properties are known to Concatentator.? */
var excludedNamespaces = [
    "http://ns.adobe.com/xap/1.0/mm/",
    "http://ns.adobe.com/xap/1.0/",
    "http://ns.adobe.com/pdf/1.3/",
    "http://ns.adobe.com/png/1.0/",
    "http://ns.adobe.com/tiff/1.0/" ];
var filteredNamespaces = {
    "http://purl.org/dc/elements/1.1/": [
        "creator", "description", "subject" ],
    "http://ns.adobe.com/photoshop/1.0/": [
        "Source", "Headline", "Credit", "DateCreated" ],
    "http://ns.adobe.com/xap/1.0/rights/": [
        "UsageTerms" ],
    "http://iptc.org/std/Iptc4xmpCore/1.0/xmlns/": [
        "CreatorContactInfo" ] };
/** What is the root (version independent part) of the AVM namespace. */
var vrNamespaceRoot = "http://www.communicatingastronomy.org/avm/";
/** The AOI keywords whose value should be a array with the same length as the
    number of exposures. */
var vrOrderedKeywords = [ "Facility", "Instrument", "Spectral.ColorAssignment",
    "Spectral.Band", "Spectral.Bandpass", "Spectral.CentralWavelength", 
    "Temporal.StartTime", "Temporal.IntegrationTime", "DatasetID",
    "FL.StretchFunction", "FL.BackgroundLevel", "FL.PeakLevel", 
    "FL.ScaledBackgroundLevel", "FL.ScaledPeakLevel", "FL.BlackLevel",
    "FL.WhiteLevel" ];
/** The AOI keywords whose value is really list(n) encoded as a rdf:Seq 
    element. */
var vrListKeywords = [ "Spatial.ReferenceValue", "Spatial.ReferenceDimension", 
    "Spatial.ReferencePixel", "Spatial.Scale", "Spatial.CDMatrix" ];

// -----------------------------------------------------------------------------
// Main
// -----------------------------------------------------------------------------

// Globals to make it easier to debug
// (stupid ExtendScript debugger doesn't show locals).
var metadata = {};           ///< Hashtable of the different metadata schemas
var xmp = "";               ///< Serialized metadata in XMP format
var composite;              ///< The final composite document
var width = 0;              ///< Width of the composite document
var height = 0;             ///< Height of the composite document
var bitdepth = 0;           ///< Bitdepth of the composite document
var resolution = Behavior.DefaultResolution;    ///< Resolution of the composite document
var sources = new Array();  ///< List of documents to merge

/** Script entry point.
    @return [undefined] */
function main() {
    // Preprocess all the documents
    for( var i = 0; i < app.documents.length; i++ ) {
        var document = app.documents[i];
        sources[i]= document;
        
        width = Math.max( width, document.width.as("px") );
        height = Math.max( height, document.height.as("px") );
        bitdepth = Math.max( bitdepth, BitsPerChannelToInt(document.bitsPerChannel) );
    }
    
    metadata = parseAll( sources );   // Parse and merge the XMP metadata
    metadata = touchUp( metadata );   // Touch up the Coverage.Spatial.Dimension field to reflect the composite size
    xmp = serializeXmp( metadata );   // Serialize the parsed XMP metadata
    composite = createDocument( width, height, resolution, bitdepth, xmp );
    mergeDocuments( composite, sources );
}

/** Converts a BitsPerChannelType to a int. 
    @param a [BitsPerChannelType] The value to be converted.
    @return [int] Converted value. */
function BitsPerChannelToInt( a ) {
    if( a == BitsPerChannelType.ONE )       return 1;
    if( a == BitsPerChannelType.EIGHT )     return 8;
    if( a == BitsPerChannelType.SIXTEEN )   return 16;
    if( a == BitsPerChannelType.THIRTYTWO ) return 32;
}

/** Converts an int to a BitsPerChannelType.
    @param a [int] The value to be converted.
    @return [BitsPerChannelType] Converted value. */
function IntToBitsPerChannel( a ) {
    switch( a ) {
        case 1 : return BitsPerChannelType.ONE;
        case 8 : return BitsPerChannelType.EIGHT;
        case 16: return BitsPerChannelType.SIXTEEN;
        case 32: return BitsPerChannelType.THIRTYTWO;
        default: return Behavior.DefaultBitDepth;
    }
}

/** Creates a new document (image) with the correct dimension and so forth. 
    Uses Behavior.CompositeName to set the name of the new document.
    @param width [int] Width of the new image.
    @param height [int] Height of the new image.
    @param resolution [int] Resolution of the new image.
    @param bitdepth [int] Bitdepth the new image.
    @param xmp [string] Serialized XMP metadata.
    @return [Document] The new image. */
function createDocument( width, height, resolution, bitdepth, xmp ) {
    var document = app.documents.add( width, height, resolution, 
        Behavior.CompositeName, NewDocumentMode.RGB, DocumentFill.TRANSPARENT );
    document.bitsPerChannel = IntToBitsPerChannel(bitdepth);
    document.xmpMetadata.rawData = xmp;
    
    return document;
}

/** Merges a sequence of documents into a single document.
    @param target [Document] Target document that should be merged into.
    @param sources [Documents] Document collection to copy pixels from.
    @return [undefined] */
function mergeDocuments( target, sources ) {
    for( var i = 0; i < sources.length; i++ ) {
        var document = sources[i];
        var name = document.name;
        
        app.activeDocument = document;
        if( Behavior.MergeFlattens )
            document.flatten();
        document.selection.selectAll();
        document.selection.copy();
        document.selection.deselect();
        
        if( Behavior.CloseDocuments && !Behavior.Debug )
            document.close( Behavior.SaveOption );
        
        app.activeDocument = target;
        target.paste();
        target.activeLayer.blendMode = BlendMode.SCREEN;
        target.activeLayer.name = name;
	}
}

/** Modifies certain VR keywords to reflect differences between individual
    exposures and the composite result.
    @param metadata [object] Reference to the shared hashtable of schemas.
    @return [object] The modified schema hashtable. */
function touchUp( metadata ) {
    // To support as many versions as possible, we search for the 
    // root of the VR namespace. And then search for each keyword we need to change
    for( var namespace in metadata ) {
        if( namespace.indexOf( vrNamespaceRoot ) == 0 ) {
            var schema = metadata[namespace];

            /*schema.keywords["Spatial.ReferenceDimension"] = { kind: Keyword.Seq, 
                value: [
                    { kind: Keyword.Simple, value: width },
                    { kind: Keyword.Simple, value: height }
                ]
            }; */
            schema.keywords["Spatial.Quality"] = null;
        }
    }

    return metadata;
}

// -----------------------------------------------------------------------------
// Data model utilities
// -----------------------------------------------------------------------------

function kindToElement( kind ) {
    switch( kind ) {
        case Keyword.Seq:
            return "Seq";
        case Keyword.Bag:
            return "Bag";
        case Keyword.Alt:
            return "Alt";
    }
    return null;
}

function elementToKind( element ) {
    if( element.getLocalName() == "Seq" )
        return Keyword.Seq;
    else if( element.getLocalName() == "Bag" )
        return Keyword.Bag;
    else if( element.getLocalName() == "Alt" )
        return Keyword.Alt;
    else
        return null;
}


// -----------------------------------------------------------------------------
// Parsing
// -----------------------------------------------------------------------------

var currentDocumentIndex;   ///< Index of the current document
var documentCount;          ///< Number of documents

/** Parses and merges metadata from a sequence of documents into a single 
    hashtable.
    @param documentArray [Documents] The documents whose metadata is to be 
    parsed and merged.
    @return [object] A hastable containing the parsed and merged metadata 
    indexed by namespace. */
function parseAll( documentArray ) {
    var xmlParser = new DOMImplementation();
    xmlParser.errorChecking = false;

    var metadata = {};
    documentCount = documentArray.length;   // documentCount is a global
    for( currentDocumentIndex = 0; currentDocumentIndex < documentCount; currentDocumentIndex++ ) {
        var xmpDocument = xmlParser.loadXML( documentArray[currentDocumentIndex].xmpMetadata.rawData );
        parseXmp( metadata, xmpDocument.getDocumentElement() );        
    }
    return metadata;
}

/** Processes a single XMP document.
    @param metadata [object] Reference to the shared hashtable of schemas.
    @param node [DOMElement] The node to process, this is expected to be a 
    x:xmpmeta element.
    @return [object] Returns the shared hashtable with new additions. */
function parseXmp( metadata, node ) {
    // TODO: assertions!

    node = node.getFirstChild();            // Get the rdf:RDF element
    
    for( var schemaNode = node.getFirstChild(); schemaNode != null;  
        schemaNode = schemaNode.getNextSibling() ) {
        
        var firstKeyword = schemaNode.getFirstChild();
        if( firstKeyword != null ) {        // Prune empty rdf:Description elements
            var namespace = firstKeyword.getNamespaceURI();
            
            // Skip the troublesome Adobe namespaces
            if( contains( namespace, excludedNamespaces ) )
                continue;

            var schema = metadata[namespace];
            if( schema == null ) {
                // Create a new schema
                schema = {
                    keywords  : {},
                    prefix    : firstKeyword.getPrefix(),
                    namespaces: {}
                };
                metadata[namespace] = schema;
            }
        }
        schema.namespaces = parseNamespaces( schema.namespaces, schemaNode );
        schema.keywords = parseSchema( schema.keywords, schemaNode, filteredNamespaces[namespace] );
    }
    
    return metadata;
}

/** Searches a node and its children for the namespaces used.
    @param namespaces [object] Namespaces used for nodes of the same time, may 
    be empty.
    @param node [DOMElement] Node to search.
    @return [object] Hashtable with the used namespaces. */
function parseNamespaces( namespaces, node ) {
    if( node.getNodeType() == DOMNode.ELEMENT_NODE ) {
        // Store the nodes namespace
        var uri = node.getNamespaceURI();
        if( uri != "http://www.w3.org/1999/02/22-rdf-syntax-ns#" && uri != "adobe:ns:meta/" )
            namespaces[node.getPrefix()] = uri;
        
        // Recurse on the children
        var child = node.getFirstChild();
        while( child != null ) {
            namespaces = parseNamespaces( namespaces, child );
            child = child.getNextSibling();
        }
    }
    return namespaces;
}

/** Processes a single schema (RDF description).
    @param keywords [object] Shared hashtable with all the keywords for this 
    schema.
    @param node [DOMElement] The node to process, this is expected to be a 
    rdf:Description element.
    @param filter [Array] An array of keyword names to include.
    @return [object] A hashtable with the keywords indexed by name. */
function parseSchema( keywords, node, filter ) {
    // TODO: assertions!

    var keywordNode = node.getFirstChild();
    for( keywordNode = node.getFirstChild(); keywordNode != null; 
        keywordNode = keywordNode.getNextSibling() ) {
        
        var name = keywordNode.getLocalName();
        if( filter != null && !contains( name, filter ) )
            continue;
            
        keywords[name] = mergeValue( 
            keywordNode,
            keywords[name], 
            parseValue( keywordNode.getFirstChild() )
        );
    }

    return keywords;
}

/** Parses a single keyword value.
    @param node [DOMNode] The keyword value.
    @return [object] If node is a string, a string is returned. If node is a 
    <rdf:Seq> an array is returned. In other cases the node is returned 
    unchanged. */
function parseValue( node ) {
    if( node != null ) {
        if( node.getNodeType() == DOMNode.TEXT_NODE ) {
            return {
                kind : Keyword.Simple,
                value: node.getNodeValue()
            };
        }
        else if( node.getNamespaceURI() == "http://www.w3.org/1999/02/22-rdf-syntax-ns#" ) {
            return {
                kind : elementToKind( node ),
                value: parseArray( node )
            };
        }
        else if( node.getNodeType() == DOMNode.ELEMENT_NODE ) {
            return {
                kind      : Keyword.Complex,
                value     : parseComplex( node )
            };
        }
    }
    return null;
}

/** Parses a complex value.
    @param node [DOMNode] Node to read from.
    @return [Hashtable] A hashtable of child properties. */
function parseComplex( node ) {
    var value = {};
    
    while( node != null ) {
        value[node.getLocalName()] = parseValue( node.getFirstChild() );
        node = node.getNextSibling();
    }
    
    return value;
}

/** Creates a list of sentinels for use with the AOI exposure lists.
    @return [object] An array of sentinels with the same number of elements as the
    number of exposures being concatenated. */
function sentinelList() {
    var list = new Array();
    for( var i = 0; i < documentCount; i++ ) {
        list[i] = { kind: Keyword.Simple, value: Behavior.Sentinel };
    }
    return list;
}

/** Parses a <rdf:Seq> node to an array.
    @param node [DOMElement] A <rdf:Seq> element.
    @return [Object] An object with properties "kind" and "values". Kind contains the kind
    of sequence (Bag or Seq) and values contains an array of values.
    Ordering is retained and each value is processed with parseValue. */
function parseArray( node ) {
    // TODO: assertions!

    var items = new Array();
    var itemCount = 0;

    var itemNode = node.getFirstChild();
    while( itemNode != null ) {
        items[itemCount++] = parseValue( itemNode.getFirstChild() );
        itemNode = itemNode.getNextSibling();
    }
    
    return items;
}

/** Determines if a list contains a string.
    @param item [String] Item to find.
    @param list [Array] list of strings to search.
    @return [bool] True if the item is in the list, false otherwise. */
function contains( item, list ) {
    for( var i = 0; i < list.length; i++ )
        if( list[i] == item )
            return true;
    return false;
}

// -----------------------------------------------------------------------------
// Merging
// -----------------------------------------------------------------------------

/** Merges two keyword values according to the merging logic.
    @param keywordNode [object] The current keyword DOM node.
    @param oldValue [object] The old (possibly semi-merged) value.
    @param newValue [object] The new value from the current document being merged. */
function mergeValue( keywordNode, oldValue, newValue ) {
    var mergedValue;    // New merged value
    
    if( keywordNode != null && keywordNode.getNamespaceURI().indexOf( vrNamespaceRoot ) == 0 ) {
        if( contains( keywordNode.getLocalName(), vrOrderedKeywords ) ) {
            // For the special AOI ordered keywords the logic goes:
            //     If the parsed value is a sequence the first item of the sequence
            //     should be inserted into the proper location in the sequence.
            if( oldValue == null ) {
                mergedValue = { kind: Keyword.Seq, value: sentinelList() };
            } else {
                mergedValue = oldValue;
            }
            
            var newItem = newValue.value[0];
            if( newItem != null ) {
                mergedValue.value[currentDocumentIndex] = newItem;
            }
        }       
        else if( contains( keywordNode.getLocalName(), vrListKeywords ) ) {
            // The Spatial.* values of type list(n) should be merged such that
            // only the last set is kept
            mergedValue = newValue;
        }
        else {
            mergedValue = mergeRegularValue( oldValue, newValue );
        }
    }
    else {
        mergedValue = mergeRegularValue( oldValue, newValue );
    }
    
    return mergedValue;
}

/** Merges a regular value (ie. not a list(n) or special AOI list).
    @param oldValue [object] Old value.
    @param newValue [object] New value.
    @return [object] Merged value. */
function mergeRegularValue( oldValue, newValue ) {
    if( oldValue != null && oldValue.kind == newValue.kind ) {
        var mergedValue = oldValue;
        switch( oldValue.kind ) {
            case Keyword.Simple:
                mergedValue.value = newValue.value;
                break;
            case Keyword.Complex:
                mergedValue.value = mergeComplex( oldValue.value, newValue.value );
                break;
            default:
                mergedValue.value = mergeArrays( oldValue.value, newValue.value );
                break;
        }
        return mergedValue;
    }
    return newValue;
}

/** Merges two arrays into one by concatenation.
    @param a [Array] First array.
    @param b [Array] Second array.
    @return [Array] Merged array. */
function mergeArrays( a, b ) {
    var c = new Array();
    for( var i = 0; i < a.length; i++ )
        c[i] = a[i];
    for( var i = 0; i < b.length; i++ )
        c[i+a.length] = b[i];
        
    return c;
}

/** Merges two sets into one set. The two sets are assumed to be proper sets
    when passed to this function.
    @param a [Array] First set.
    @param b [Array] Second set.
    @return [Array] Merged set. */
function mergeSets( a, b ) {
    var c = new Array();
    
    if( a.length > b.length ) {
        for( var i = 0; i < a.length; i++ )
            c[i] = a[i];

        for( var i = 0, j = 0; i < b.length; i++ )
            if( !contains( b[i], c ) )  // Doesn't work
                c[a.length + j++] = b[i];
    }
    else {
        for( var i = 0; i < b.length; i++ )
            c[i] = b[i];

        for( var i = 0, j = 0; i < a.length; i++ )
            if( !contains( a[i], c ) )  // Doesn't work
                c[b.length + j++] = a[i];    
    }
    
    return c;
}

/** Merges a complex value.
    @param oldValue [Hashtable] Hashtable of old values
    @param newValue [Hashtable] Hashtable of new values
    @return [Hashtable] Merged value. */
function mergeComplex( oldValue, newValue ) {
    var mergedValue = {};
    for( var name in oldValue ) {
        mergedValue[name] = mergeValue( null, oldValue[name], newValue[name] );
    }
    return mergedValue;
}

// -----------------------------------------------------------------------------
// Serialization
// -----------------------------------------------------------------------------

/** Serializes the schema hashtable to a string.
    @param metadata [object] The schemas to serialize.
    @return [string] The serialized XMP metadata document. */
function serializeXmp( metadata ) {
    var xmp = "";
    
    xmp = xmp + '<?xpacket begin="?" id="W5M0MpCehiHzreSzNTczkc9d"?>\n';
    xmp = xmp + '<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="3.1.1-111">\n <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n';

    for( var namespace in metadata ) {
        xmp = xmp + serializeSchema( metadata[namespace] );
    }
    
    xmp = xmp + ' </rdf:RDF>\n</x:xmpmeta>\n';
    xmp = xmp + '<?xpacket end="w"?>';
    return xmp;
}

/** Serializes a single schema.
    @param schema [object] A keyword schema hastable.
    @return [string] The serialized XMP metadata schema. */
function serializeSchema( schema ) {
    var xml = "";
    
    xml = xml + '  <rdf:Description rdf:about=""';
    for( var prefix in schema.namespaces )
        if( prefix.length > 0 )
            xml = xml + ' xmlns:' + prefix + '="' + schema.namespaces[prefix] + '"';
    xml = xml + '>\n';

    for( var name in schema.keywords ) {
        xml = xml + serializeKeyword( schema.prefix, name, schema.keywords[name] );
    }
    xml = xml + '  </rdf:Description>\n';
    
    return xml;
}

/** Serializes a single keyword.
    @param prefix [string] Namespace prefix.
    @param name [string] Keyword name.
    @param keyword [object] Keyword (kind, value) tuple.
    @return [string] The serialized keyword. */
function serializeKeyword( prefix, name, keyword ) {
    var xml = "";
    
    if( keyword != null ) {
        xml = xml + '    <' + prefix + ':' + name;
        if( keyword.kind == Keyword.Complex )
            xml = xml + ' rdf:parseType="Resource"';
        xml = xml + '>';
		xml = xml + serializeValue( prefix, keyword );
        if( keyword.kind == Keyword.Complex )
            xml = xml + '    ';
        xml = xml + '</' + prefix + ':' + name + '>\n';
    }
    
    return xml;
}

/** Serialize a keyword value.
    @params prefix [string] Namespace prefix.
    @paramm keyword [object] Keyword (kind,value) tuple.
    @return [string] The serialized value. */
function serializeValue( prefix, keyword ) {
    switch( keyword.kind ) {
    case Keyword.Simple:
        return keyword.value;
    case Keyword.Complex:
        return serializeComplex( prefix, keyword.value );
    default:
        return serializeArray( prefix, keyword.kind, keyword.value );
    }
}

/** Serializes an array to a <rdf:Seq> or <rdf:Bag> node.
    @param kind [int] The kind of sequence.
    @param value [Array] Array of values.
    @return [string] The serialized array. */
function serializeArray( prefix, kind, value ) {
    if( value != null && value.length > 0 ) {
        var xml = '\n';
        
        if( kind == Keyword.Alt ) {
            xml = xml + '     <rdf:Alt>\n';
            xml = xml + '       <rdf:li xml:lang="x-default">';
			xml = xml + serializeValue( prefix, value[0] );
            xml = xml + '</rdf:li>\n';
            xml = xml + '     </rdf:Alt>\n';
        } else {
            var kindElement = kindToElement( kind );
        
            xml = xml + '     <rdf:' + kindElement + '>\n';
            for( var i = 0; i < value.length; i++ )
			{
				xml = xml + '      <rdf:li>' + serializeValue( prefix, value[i] ) + '</rdf:li>\n';
            }
			xml = xml + '     </rdf:' + kindElement + '>\n    ';
        }
        
        return xml;
    }
    return "";
}

function serializeComplex( prefix, value ) {
    var xml = "";
	for( var name in value ) {
		if ( value[name] != null)
		{
			xml = xml + '\n     <' + prefix + ':' + name + '>';		
			xml = xml + serializeValue( prefix, value[name] );
			xml = xml + '</' + prefix + ':' + name + '>';
		}
	}

    return xml + '\n';
}

// -----------------------------------------------------------------------------
// User Interface
// -----------------------------------------------------------------------------

/** Displays a confirmation dialog to the user.
    @return [bool] True if the process is to continue, false if the user wishes 
    to cancel the process. */
function confirmationDialog() {
	if( !Behavior.CloseDocuments )
		return true;
    return confirm( 
        "The ESA/ESO/NASA FITS Concatenator combines the metadata from two or "
        + "more individual exposures. Each exposure is copied to a layer in "
        + "the composite image and the original file is closed, is this ok?",
        true, "The ESA/ESO/NASA FITS Concatenator" );
}

// -----------------------------------------------------------------------------
// Script bootstrapper
// -----------------------------------------------------------------------------

if( Behavior.Debug ) {
    if( app.documents.length > 1 ) {
        main();
    }
} else {
    try {
        // Make sure there is anything to do
        if( app.documents.length > 1 ) {
            if( confirmationDialog() ) {
                main();
            }
        }
    }
    catch(e) {
        alert( "An unexpected error occurred, please notify the development team"
                + " at http://spacetelescope.org/fits_liberator/" );
    }
}
